﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

local TbAutoTracker = LibStub("AceAddon-3.0"):NewAddon("TbAutoTracker", "AceEvent-3.0")
local TbCore = TbCore

local MAX_BATTLEFIELD_QUEUES = MAX_BATTLEFIELD_QUEUES
local GetBattlefieldStatus = GetBattlefieldStatus
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitAffectingCombat = UnitAffectingCombat
local UnitClass = UnitClass
local SpellIsTargeting = SpellIsTargeting
local SetTracking = SetTracking
local GetNumTrackingTypes = GetNumTrackingTypes
local GetTrackingInfo = GetTrackingInfo

local _teamSize = 0
local _stealthers = 0
local _trackHidden = GetSpellInfo(19885)
local _trackHumanoids = GetSpellInfo(19883)
local _improved =
{
    [GetSpellInfo(1494)] = true,    -- Track Beasts.
    [GetSpellInfo(19878)] = true,   -- Track Demons.
    [GetSpellInfo(19879)] = true,   -- Track Dragonkin.
    [GetSpellInfo(19880)] = true,   -- Track Elementals.
    [GetSpellInfo(19882)] = true,   -- Track Giants.
    [_trackHumanoids] = true,       -- Track Humanoids.
    [GetSpellInfo(19884)] = true,   -- Track Undead.
}


-------------------------------------------------------------------------------
-- Initialization.
-------------------------------------------------------------------------------
function TbAutoTracker:OnEnable()
    --
    -- Only enable this addon for hunters of the appropriate level.
    --
    local _, class = UnitClass("player")
    if (class ~= "HUNTER" or UnitLevel("player") < 10) then return end

    --
    -- Check our tracking. This is done here in case we are reloading during arena preparation, but would be redundant with the check performed
    -- upon PLAYER_ALIVE if we are initially loading the game.
    --
    self:PLAYER_ALIVE()

    --
    -- Check where we are after seeing a loading screen or after being resurrected.
    --
    self:RegisterEvent("PLAYER_ALIVE")
    self:RegisterEvent("PLAYER_UNGHOST", "PLAYER_ALIVE")
    self:RegisterEvent("CHARACTER_POINTS_CHANGED", "PLAYER_ALIVE")
    self:RegisterEvent("PLAYER_TALENT_UPDATE", "PLAYER_ALIVE")
    self:RegisterEvent("ARENA_OPPONENT_UPDATE")
end


-------------------------------------------------------------------------------
-- Ensure that we have the correct tracking type.
-------------------------------------------------------------------------------
function TbAutoTracker:PLAYER_ALIVE(e)
    if (self:ShouldNotSetTracking()) then return end

    local current, desired = self:GetCurrentTracking()
    if (TbCore:IsInArena()) then
        --
        -- Check how many opponents in total are in the opposing team. If we were previous unaware of the size of the opposing
        -- team, we might have to recheck all of its members.
        --
        local checkAll = _teamSize == 0
        self:UpdateNumExpectedArenaOpponents()

        local hidden, humanoids = TbAutoTracker:GetPvPTrackingIndices()
        if (self:HasImprovedTracking()) then
            --
            -- Count the number of stealthers on the other team if this is the first time we enter this method this match.
            -- If we've just entered a match, then the following loop will not do anything. Otherwise, we must be here
            -- due to a UI reload, in which case we need to re-evaluate the current match.
            --
            if (checkAll) then
                for i = 1, GetNumArenaOpponents() do
                    local unit = "arena" .. i
                    local _, class = UnitClass(unit)
                    if (class == "DRUID" or class == "ROGUE") then
                        _stealthers = _stealthers + 1
                    end
                end
            end

            --
            -- Set the appropriate tracking type based on the number of stealthers on the opposing team.
            --
            if (_stealthers == 0) then
                if (current ~= _trackHidden) then
                    SetTracking(hidden)
                end
            elseif (current ~= _trackHumanoids) then
                SetTracking(humanoids)
            end
        elseif (current ~= _trackHidden) then
            SetTracking(hidden)
        end
    else
        --
        -- Since we are no longer in an arena, reset our arena variables.
        --
        _teamSize = 0
        _stealthers = 0
        if (TbCore:IsInBattleground() or TbCore:IsInActiveWintergrasp()) then
            --
            -- Use Track Hidden in battlegrounds, unless the player already had an improved form of tracking active, in which
            -- case use Track Humanoids.
            --
            local hidden, humanoids = TbAutoTracker:GetPvPTrackingIndices()
            if (self:HasImprovedTracking() and _improved[current]) then
                if (current ~= _trackHumanoids) then
                    SetTracking(humanoids)
                end
            elseif (current ~= _trackHidden) then
                SetTracking(hidden)
            end
        elseif (TbCore:IsInPvEInstance() and self:HasImprovedTracking()) then
            --
            -- Read our tracking and change our tracking to the first available improved tracking type, if necessary.
            --
            if (not _improved[current]) then
                SetTracking(desired)
            end
        end
    end
end

function TbAutoTracker:ARENA_OPPONENT_UPDATE(e, unit, kind)
    --
    -- Abort if we are not discovering a new unit.
    --
    if (kind ~= "seen" or strfind(unit, "pet$") or not self:HasImprovedTracking() or self:ShouldNotSetTracking()) then return end

    --
    -- Check whether this is a stealther.
    --
    local _, class = UnitClass(unit)
    if (class == "DRUID" or class == "ROGUE") then
        _stealthers = _stealthers + 1
    end
    if (_stealthers > 0) then return end

    --
    -- Check how many opponents in total are on the opposing team. If there are no stealthers, use Track Humanoids.
    --
    self:UpdateNumExpectedArenaOpponents()
    if (GetNumArenaOpponents() == _teamSize and _stealthers == 0 and self:HasImprovedTracking() and self:GetCurrentTracking() ~= _trackHumanoids) then
        local hidden, humanoids = TbAutoTracker:GetPvPTrackingIndices()
        SetTracking(humanoids)
    end
end


-------------------------------------------------------------------------------
-- Helper methods.
-------------------------------------------------------------------------------
function TbAutoTracker:UpdateNumExpectedArenaOpponents()
    --
    -- Get the expected number of opponents in this arena match.
    --
    if (_teamSize > 0) then return end
    for i = 1, MAX_BATTLEFIELD_QUEUES do
        local status, _, _, _, _, teamSize = GetBattlefieldStatus(i)
        if (status == "active") then
            if (teamSize == _teamSize) then return end
            _teamSize = teamSize
            break
        end
    end
end

function TbAutoTracker:ShouldNotSetTracking()
    return UnitAffectingCombat("player") or UnitIsDeadOrGhost("player") or SpellIsTargeting() or (TbCore:IsInArena() and GetBattlefieldWinner())
end


-------------------------------------------------------------------------------
-- Tracking helper methods.
-------------------------------------------------------------------------------
function TbAutoTracker:GetCurrentTracking()
    local currentName
    local desiredIndex = 0
    for i = 1, GetNumTrackingTypes() do
        --
        -- Check trackings sequentially. If this is an improved tracking type, save its index.
        --
        local name, _, active = GetTrackingInfo(i)
        if (_improved[name]) then
            desiredIndex = i
        end

        --
        -- If this is the active tracking type, then save its name.
        --
        if (active) then
            currentName = name
        end

        --
        -- Once we have determined our current tracking and the index of the first improved tracking type, we are done.
        --
        if (currentName and desiredIndex > 0) then break end
    end

    --
    -- Return the name of the current tracking type and the index of the first improvied tracking type.
    --
    return currentName, desiredIndex
end

function TbAutoTracker:GetPvPTrackingIndices()
    local hidden = 0
    local humanoids = 0
    for i = 1, GetNumTrackingTypes() do
        --
        -- Check trackings sequentially to find Track Hidden and Track Humanoids
        --
        local name = GetTrackingInfo(i)
        if (name == _trackHidden) then hidden = i
        elseif (name == _trackHumanoids) then humanoids = i end

        --
        -- Once we have determined the indices corresponding to Track Hidden and Track Humanoids, we are done.
        --
        if (hidden > 0 and humanoids > 0) then break end
    end

    --
    -- Return the indices of Track Hidden and Track Humanoids.
    --
    return hidden, humanoids
end


function TbAutoTracker:HasImprovedTracking()
    return TbCore:HasTalent(3, 1)
end
